import copy

from dune.perftool.loopy.symbolic import substitute
from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           name_coefficientcontainer,
                                           pymbolic_coefficient,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      barrier,
                                      built_instruction,
                                      domain,
                                      function_mangler,
                                      generator_factory,
                                      get_counter,
                                      get_global_context_value,
                                      globalarg,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      retrieve_cache_items,
                                      silenced_warning,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import (get_buffer_temporary,
                                        initialize_buffer,
                                        switch_base_storage,
                                        )
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.spaces import (name_lfs,
                                         name_lfs_bound,
                                         )
from dune.perftool.pdelab.geometry import (local_dimension,
                                           world_dimension,
                                           )
from dune.perftool.sumfact.amatrix import (AMatrix,
                                           LargeAMatrix,
                                           quadrature_points_per_direction,
                                           basis_functions_per_direction,
                                           construct_amatrix_sequence,
                                           )
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.loopy.symbolic import SumfactKernel
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.error import PerftoolError
from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )
from dune.perftool.sumfact.quadrature import quadrature_inames
from dune.perftool.sumfact.vectorization import find_sumfact
from loopy.symbolic import FunctionIdentifier, IdentityMapper

import loopy as lp
import numpy as np
import pymbolic.primitives as prim
from pytools import product


@function_mangler
def horizontal_add_function_mangler(knl, func, arg_dtypes):
    if func == "horizontal_add":
        from dune.perftool.loopy.vcl import get_vcl_type
        vcl = lp.types.NumpyType(get_vcl_type(np.float64, register_size=256))
        return lp.CallMangleInfo("horizontal_add", (lp.types.NumpyType(np.float64),), (vcl,))


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


@kernel_cached
def accum_iname(restriction, bound, i):
    return sumfact_iname(bound, "accum")


def setup_theta(inp, element, restriction, component, index):
    if index is None:
        index = ()
    else:
        index = (index,)
    # Write initial coefficients into buffer
    lfs = name_lfs(element, restriction, component)
    basisiname = sumfact_iname(name_lfs_bound(lfs), "basis")
    container = name_coefficientcontainer(restriction)
    coeff = pymbolic_coefficient(container, lfs, basisiname)
    assignee = Subscript(Variable(inp), (Variable(basisiname),) + index)
    return instruction(assignee=assignee,
                       expression=coeff,
                       )


def name_test_function_contribution(test):
    count = get_counter('sumfact_contribution_counter')
    grad = "grad_" if test.reference_grad else ""
    return restricted_name("contrib_{}phi_{}".format(grad, str(count)), test.restriction)


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    dim = world_dimension()

    # If this is a gradient, we find the gradient iname
    additional_inames = frozenset({})
    if accterm.argument.index:
        for i in accterm.argument.index._indices:
            if i not in visitor.dimension_indices:
                from dune.perftool.pdelab.localoperator import grad_iname
                additional_inames = additional_inames.union(frozenset({grad_iname(i, dim)}))

    def emit_sumfact_kernel(i, restriction, insn_dep):
        # Construct the matrix sequence for this sum factorization
        a_matrices = construct_amatrix_sequence(transpose=True,
                                                derivative=i if accterm.argument.index else None,
                                                facedir=get_facedir(accterm.argument.restriction),
                                                facemod=get_facemod(accterm.argument.restriction),
                                                )

        # Get the vectorization info. If this happens during the dry run, we get dummies
        from dune.perftool.sumfact.vectorization import get_vectorization_info
        a_matrices, buf, inp, index, padding = get_vectorization_info(a_matrices, (accterm.argument.restriction, restriction))

        # Initialize a base storage for this buffer and get a temporay pointing to it
        shape = tuple(mat.cols for mat in a_matrices if mat.cols != 1)
        dim_tags = ",".join(['f'] * local_dimension())
        if index is not None:
            shape = shape + (4,)
            dim_tags = dim_tags + ",c"
            index = (index,)
            vectag = frozenset({"gradvec"})
        else:
            index = ()
            vectag = frozenset()

        temp = initialize_buffer(buf,
                                 base_storage_size=product(max(mat.rows, mat.cols) for mat in a_matrices),
                                 num=2
                                 ).get_temporary(shape=shape,
                                                 dim_tags=dim_tags,
                                                 name=inp,
                                                 )

        # Those input fields, that are padded need to be set to zero in order to do a horizontal_add lateron
        for pad in padding:
            instruction(assignee=prim.Subscript(prim.Variable(temp), tuple(Variable(i) for i in quadrature_inames()) + (pad,)),
                        expression=0,
                        forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                        forced_iname_deps_is_final=True,
                        tags=frozenset(["quadvec"])
                        )

        # Replace gradient iname with correct index for assignement
        replace_dict = {}
        for iname in additional_inames:
            replace_dict[Variable(iname)] = i
        expression = substitute(pymbolic_expr, replace_dict)

        # Determine dependencies
        from loopy.match import Or, Writes
        from loopy.symbolic import DependencyMapper
        from dune.perftool.tools import get_pymbolic_basename
        deps = Or(tuple(Writes(get_pymbolic_basename(expr)) for expr in DependencyMapper()(expression)))

        # Issue an instruction in the quadrature loop that fills the buffer
        # with the evaluation of the contribution at all quadrature points
        assignee = Subscript(Variable(temp), tuple(Variable(i) for i in quadrature_inames()) + index)
        contrib_dep = instruction(assignee=assignee,
                                  expression=expression,
                                  forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                                  forced_iname_deps_is_final=True,
                                  tags=frozenset({"quadvec"}).union(vectag),
                                  depends_on=frozenset({deps})
                                  )

        if insn_dep is None:
            insn_dep = frozenset({contrib_dep})

        # Add a sum factorization kernel that implements the multiplication
        # with the test function (stage 3)
        pref_pos = i if accterm.argument.index else None
        result, insn_dep = sum_factorization_kernel(a_matrices,
                                                    buf,
                                                    3,
                                                    insn_dep=insn_dep,
                                                    additional_inames=frozenset(visitor.inames),
                                                    preferred_position=pref_pos,
                                                    restriction=(accterm.argument.restriction, restriction),
                                                    )

        inames = tuple(accum_iname((accterm.argument.restriction, restriction), mat.rows, i) for i, mat in enumerate(a_matrices))

        # Collect the lfs and lfs indices for the accumulate call
        test_lfs = determine_accumulation_space(accterm.argument.expr, 0, measure)
        test_lfs.index = flatten_index(tuple(Variable(i) for i in inames),
                                       (basis_functions_per_direction(),) * dim,
                                       order="f"
                                       )

        # In the jacobian case, also determine the space for the ansatz space
        ansatz_lfs = determine_accumulation_space(accterm.term, 1, measure)
        rank = 2 if visitor.inames else 1
        if rank == 2:
            # TODO the next line should get its inames from elsewhere. This is *NOT* robust (but works right now)
            ansatz_lfs.index = flatten_index(tuple(Variable(visitor.inames[i]) for i in range(world_dimension())),
                                             (basis_functions_per_direction(),) * dim,
                                             order="f"
                                             )

        # Construct the expression representing "{r,jac}.accumulate(..)"
        accum = name_accumulation_variable(ansatz_lfs.get_restriction() + test_lfs.get_restriction())

        # Determine the expression to accumulate with. This depends on the vectorization strategy!
        result = prim.Subscript(result, tuple(prim.Variable(i) for i in inames))
        vecinames = ()
        if index:
            iname = accum_iname((accterm.argument.restriction, restriction), 4, "vec")
            vecinames = (iname,)
            transform(lp.tag_inames, [(iname, "vec")])
            from dune.perftool.tools import maybe_wrap_subscript
            result = prim.Call(prim.Variable("horizontal_add"),
                               (maybe_wrap_subscript(result, prim.Variable(iname)),),
                               )

        # In the case of FastDGGridOperator we can write directly into the resiudal/jacobi
        #
        # TODO: At the moment this only works if we do not vectorize
        # (over gradients) because loopy tries to acces a vectorclass
        # variable.
        if get_option('fastdg'):
            ft = get_global_context_value("form_type")
            if ft=='residual':
                accum = accum + ".data()"
                size = basis_functions_per_direction() ** world_dimension()
                globalarg(accum, dtype=np.float64, shape=(size,), managed=False)
                assignee = prim.Subscript(prim.Variable(accum), (test_lfs.index,))
                expression = prim.Sum((assignee,result))
                instruction(assignee=assignee,
                            expression=expression,
                            forced_iname_deps=frozenset(inames),
                            forced_iname_deps_is_final=True,
                            depends_on=insn_dep,
                            )
            else:
                assert ft=='jacobian'
                accum = accum + ".data()"
                size = basis_functions_per_direction() ** world_dimension()
                globalarg(accum, dtype=np.float64, shape=(size, size), managed=True)
                assignee = prim.Subscript(prim.Variable(accum), (ansatz_lfs.index, test_lfs.index))
                expression = prim.Sum((assignee,result))
                instruction(assignee=assignee,
                            expression=expression,
                            forced_iname_deps=frozenset(inames + visitor.inames),
                            forced_iname_deps_is_final=True,
                            depends_on=insn_dep,
                            )
        # Default: Generate accumulation instructions
        else:
            expr = Call(PDELabAccumulationFunction(accum, rank),
                        (ansatz_lfs.get_args() +
                         test_lfs.get_args() +
                         (result,)
                        )
            )
            instruction(assignees=(),
                        expression=expr,
                        forced_iname_deps=frozenset(inames + visitor.inames + vecinames),
                        forced_iname_deps_is_final=True,
                        depends_on=insn_dep,
            )


        # Mark the transformation that moves the quadrature loop inside the trialfunction loops for application
        transform(nest_quadrature_loops, visitor.inames)

        return insn_dep

    # Extract the restrictions on argument-1:
    jac_restrictions = frozenset(tuple(ma.restriction for ma in extract_modified_arguments(accterm.term, argnumber=1)))
    if not jac_restrictions:
        jac_restrictions = frozenset({0})

    insn_dep = None
    for restriction in jac_restrictions:
        if accterm.argument.index:
            for i in range(world_dimension()):
                insn_dep = emit_sumfact_kernel(i, restriction, insn_dep)
        else:
            insn_dep = emit_sumfact_kernel(None, restriction, insn_dep)


@generator_factory(item_tags=("sumfactkernel",), context_tags=("kernel",), cache_key_generator=lambda a, b, s, **kw: (a, b, s, kw.get("restriction", 0)))
def sum_factorization_kernel(a_matrices,
                             buf,
                             stage,
                             insn_dep=frozenset({}),
                             additional_inames=frozenset({}),
                             preferred_position=None,
                             outshape=None,
                             restriction=0,
                             direct_input=None,
                             ):
    """
    Calculate a sum factorization matrix product.

    Y = A_{d-1}*...*A_0*X

    where X is the input matrix and Y is the output variable.

    Arguments:
    ----------
    a_matrices: An iterable of AMatrix instances
        The list of tensors to be applied to the input.
        Order of application is from 0 up.
    buf: A string identifying the flip flop buffer in use
        for intermediate results. The memory is expected to be
        pre-initialized with the input.
    insn_dep: an instruction ID that the first issued instruction
        should depend upon. All following ones will depend on each
        other.
    """
    if get_global_context_value("dry_run", False):
        return SumfactKernel(a_matrices, buf, stage, preferred_position, restriction), frozenset()

    ftags = "f,f"
    ctags = "c,c"
    vec_shape = ()
    if next(iter(a_matrices)).vectorized:
        ftags = ftags + ",vec"
        ctags = ctags + ",vec"
        vec_shape = (4,)

    insn_dep = frozenset({barrier(depends_on=insn_dep,
                                  within_inames=additional_inames,
                                  )})

    for l, a_matrix in enumerate(a_matrices):
        # Get a temporary that interprets the base storage of the input
        # as a column-major matrix. In later iteration of the amatrix loop
        # this reinterprets the output of the previous iteration.
        inp_shape = (a_matrix.cols, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))

        # TODO: fastdg and vectorization
        if l==0 and direct_input is not None and not a_matrix.vectorized:
            globalarg(direct_input, dtype=np.float64, shape=inp_shape)
            inp = direct_input
        else:
            inp = get_buffer_temporary(buf,
                                       shape=inp_shape + vec_shape,
                                       dim_tags=ftags)

        # The input temporary will only be read from, so we need to silence the loopy warning
        silenced_warning('read_no_write({})'.format(inp))

        switch_base_storage(buf)

        # Get a temporary that interprets the base storage of the output
        # as row-major matrix
        out_shape = (a_matrix.rows, product(mat.rows for mat in a_matrices[:l]) * product(mat.cols for mat in a_matrices[l + 1:]))

        out = get_buffer_temporary(buf,
                                   shape=out_shape + vec_shape,
                                   dim_tags=ctags)

        # Get the inames needed for one matrix-matrix multiplication
        i = sumfact_iname(out_shape[0], "row")
        j = sumfact_iname(out_shape[1], "col")

        # Maybe introduce a vectorization iname for this matrix-matrix multiplication
        vec_iname = ()
        if a_matrix.vectorized:
            iname = sumfact_iname(4, "vec")
            vec_iname = (prim.Variable(iname),)
            transform(lp.tag_inames, [(iname, "vec")])

        if a_matrix.cols == 1:
            # A trivial reduction is implemented as a product, otherwise we run into
            # a code generation corner case producing way too complicated code. This
            # could be fixed upstream, but the loopy code realizing reductions is not
            # trivial and the priority is kind of low.
            matprod = Product((prim.Subscript(prim.Variable(a_matrix.name), (prim.Variable(i), 0) + vec_iname),
                               prim.Subscript(prim.Variable(inp), (0, prim.Variable(j)) + vec_iname)
                               ))
        else:
            k = sumfact_iname(a_matrix.cols, "red")

            # Construct the matrix-matrix-multiplication expression a_ik*in_kj
            prod = prim.Product((prim.Subscript(prim.Variable(a_matrix.name),
                                                (prim.Variable(i), prim.Variable(k)) + vec_iname),
                                 prim.Subscript(prim.Variable(inp),
                                                (prim.Variable(k), prim.Variable(j)) + vec_iname)
                                ))
            matprod = lp.Reduction("sum", k, prod)

        # Issue the reduction instruction that implements the multiplication
        # at the same time store the instruction ID for the next instruction to depend on
        assignee = prim.Subscript(prim.Variable(out), (prim.Variable(i), prim.Variable(j)) + vec_iname)
        insn_dep = frozenset({instruction(assignee=assignee,
                                          expression=matprod,
                                          forced_iname_deps=frozenset({i, j}).union(additional_inames),
                                          forced_iname_deps_is_final=True,
                                          depends_on=insn_dep,
                                          )
                              })

    if outshape is None:
        outshape = tuple(mat.rows for mat in a_matrices if mat.rows != 1)
    dim_tags = ",".join(['f'] * len(outshape))

    if next(iter(a_matrices)).vectorized:
        outshape = outshape + vec_shape
        # This is a 'bit' hacky: In stage 3 we need to return something with vectag, in stage 1 not.
        if stage == 1:
            dim_tags = dim_tags + ",c"
        else:
            dim_tags = dim_tags + ",vec"

    out = get_buffer_temporary(buf,
                               shape=outshape,
                               dim_tags=dim_tags,
                               )
    silenced_warning('read_no_write({})'.format(out))

    return next(iter(a_matrices)).output_to_pymbolic(out), insn_dep
